from __future__ import annotations

from simudator.core.signal import Signal


class MiaBusConnector:
    """
    Inherited by modules that are connected to a bus in the MIA archtecture.
    Has logic for controlling when to read and write to bus.
    """

    def __init__(self, bus_control: Signal, bus_id: int = 0) -> None:
        # init the name
        self.bus_id = bus_id
        self.bus_control_s = bus_control
        # set the module as destination of the control signal
        self.bus_control_s.add_destination(self)

    def read_from_bus(self):
        bus_instruction = self.bus_control_s.get_value()
        if bus_instruction is None:
            return
        return bus_instruction[1] == self.bus_id

    def write_to_bus(self):
        bus_instruction = self.bus_control_s.get_value()
        if bus_instruction is None:
            return
        return bus_instruction[0] == self.bus_id

    def get_input_signals(self) -> list[Signal]:
        return [self.bus_control_s]

    def get_output_signals(self) -> list[Signal]:
        return []
