from simudator.core.processor import Processor
from simudator.core.signal import Signal
from simudator.processor.mia.modules.alu import ALU

cpu = Processor()
input_signal_a = Signal(cpu)
input_signal_b = Signal(cpu)
input_signal_hr = Signal(cpu)
control_signal = Signal(cpu)
output_signal = Signal(cpu)
hr_signal = Signal(cpu)
z_signal = Signal(cpu)
n_signal = Signal(cpu)
o_signal = Signal(cpu)
c_signal = Signal(cpu)
alu = ALU(
    "alu",
    input_signal_a,
    input_signal_b,
    input_signal_hr,
    control_signal,
    output_signal,
    hr_signal,
    z_signal,
    n_signal,
    o_signal,
    c_signal,
)
cpu.add_module(alu)


def test_int_binary_converter():
    # test int to binary
    assert alu.int_to_bin(0) == [0 for i in range(alu.WORD_LENGTH)]
    assert alu.int_to_bin(1) == [0] * (alu.WORD_LENGTH - 1) + [1]
    assert alu.int_to_bin(32767) == [0] + [1] * (alu.WORD_LENGTH - 1)
    assert alu.int_to_bin(9) == [0] * (alu.WORD_LENGTH - 4) + [1, 0, 0, 1]

    # test binary to int
    assert alu.bin_to_int(alu.int_to_bin(15)) == 15
    assert alu.bin_to_int(alu.int_to_bin(0)) == 0
    assert alu.bin_to_int(alu.int_to_bin(2**16)) == 2**16 - 1
    assert alu.bin_to_int(alu.int_to_bin(3576)) == 3576
    assert alu.bin_to_int([0 for i in range(alu.WORD_LENGTH)]) == 0
    assert alu.bin_to_int([0] * (alu.WORD_LENGTH - 1) + [1]) == 1
    assert alu.bin_to_int([1 for i in range(alu.WORD_LENGTH)]) == 2**16 - 1
    assert alu.bin_to_int([0] + [1] * (alu.WORD_LENGTH - 1)) == 32767
    assert alu.bin_to_int([0] * (alu.WORD_LENGTH - 4) + [1, 0, 0, 1]) == 9


def test_ones_complement():
    # test bitwise not
    assert alu.ones_complement(0) == 2**16 - 1
    assert alu.ones_complement(1) == 2**16 - 1 - 1
    assert alu.ones_complement(2) == 2**16 - 1 - 2
    assert alu.ones_complement(300) == 2**16 - 1 - 300


def test_negative_bin():
    assert alu.bin_to_int_twocomp(alu.int_to_negative_bin(0)) == 0
    assert alu.bin_to_int_twocomp(alu.int_to_negative_bin(10)) == -10
    assert alu.bin_to_int_twocomp(alu.int_to_negative_bin(1337)) == -1337
    assert alu.bin_to_int_twocomp(alu.int_to_negative_bin(2**15 - 1)) == -(
        2**15 - 1
    )
    assert alu.bin_to_int_twocomp(alu.int_to_negative_bin(2**16 - 1)) == 1
    assert alu.bin_to_int_twocomp(alu.int_to_negative_bin(2**16 - 10)) == 10
    assert alu.bin_to_int_twocomp(alu.int_to_negative_bin(2**16 - 4567)) == 4567


def test_add_numbers():
    # test binary sum
    assert alu.add_numbers(1, 1, False) == 2

    assert alu.add_numbers(1, 1, True) == 2
    assert alu.z_value == 0
    assert alu.n_value == 0
    assert alu.o_value == 0
    assert alu.c_value == 0

    assert alu.add_numbers(14, 20, False) == 34

    assert alu.add_numbers(0, 0, True) == 0
    assert alu.z_value == 1
    assert alu.n_value == 0
    assert alu.o_value == 0
    assert alu.c_value == 0

    assert alu.add_numbers(1, 2**16 - 1, True) == 0
    assert alu.z_value == 1
    assert alu.n_value == 0
    assert alu.o_value == 0
    assert alu.c_value == 1


def test_binary_diff():
    assert alu.subtract_numbers(1, 1, True) == 0
    assert alu.z_value == 1
    assert alu.n_value == 0
    assert alu.o_value == 0
    assert alu.c_value == 1
    assert alu.subtract_numbers(0, 1, True) == 65535
    assert alu.z_value == 0
    assert alu.n_value == 1
    assert alu.o_value == 0
    assert alu.c_value == 0
    assert alu.subtract_numbers(0xA100, 0xA010, True) == (0xA100 - 0xA010)
    assert alu.z_value == 0
    assert alu.n_value == 0
    assert alu.o_value == 0
    assert alu.c_value == 1
    assert alu.subtract_numbers(0xA010, 0xA100, True) == 0xFFFF + 1 + (0xA010 - 0xA100)
    assert alu.z_value == 0
    assert alu.n_value == 1
    assert alu.o_value == 0
    assert alu.c_value == 0
    assert alu.subtract_numbers(30, 29, False) == 1
    assert alu.subtract_numbers(10, 3, False) == 7
    assert alu.subtract_numbers(4, 2, False) == 2
    assert alu.subtract_numbers(500, 2, False) == 498
    assert alu.subtract_numbers(10, 2, False) == 8


def test_and_numbers():
    assert alu.and_numbers(1, 2) == 0
    assert alu.and_numbers(1, 1) == 1
    assert alu.and_numbers(3, 1) == 1
    assert alu.and_numbers(3, 2) == 2
    assert alu.and_numbers(0, 0) == 0
    assert alu.and_numbers(20, 0) == 0
    assert alu.and_numbers(5, 0) == 0


def test_or_numbers():
    assert alu.or_numbers(1, 2) == 3
    assert alu.or_numbers(1, 1) == 1
    assert alu.or_numbers(0, 0) == 0
    assert alu.or_numbers(20, 0) == 20


def test_binary_logical_left_shift():
    assert alu.logical_left_shift(0b01) == 0b10
    assert alu.c_value == 0
    assert alu.logical_left_shift(0b11) == 0b110
    assert alu.c_value == 0
    assert alu.logical_left_shift(0b0) == 0b0
    assert alu.c_value == 0
    assert alu.logical_left_shift(0b1000000000000000) == 0b0
    assert alu.c_value == 1


def test_arithmetic_right_shift():
    assert alu.arithmetic_right_shift(0) == 0
    assert alu.c_value == 0
    assert alu.arithmetic_right_shift(1) == 0
    assert alu.c_value == 1
    assert alu.arithmetic_right_shift(2) == 1
    assert alu.c_value == 0
    assert alu.arithmetic_right_shift(0b1111111111111111) == 0b1111111111111111
    assert alu.c_value == 1
    assert alu.arithmetic_right_shift(0b0111111111111111) == 0b0011111111111111
    assert alu.c_value == 1


def test_logical_right_shift():
    assert alu.logical_right_shift(0) == 0
    assert alu.c_value == 0
    assert alu.logical_right_shift(1) == 0
    assert alu.c_value == 1
    assert alu.logical_right_shift(2) == 1
    assert alu.c_value == 0


def test_rotate_left():
    assert alu.left_rotate(0) == 0
    assert alu.c_value == 0
    assert alu.left_rotate(1) == 2


def test_signals():
    # Test just out
    control_signal.update_value(1)
    input_signal_b.update_value(15)
    cpu.do_tick()
    assert output_signal.get_value() == 15

    # Test do nothing
    control_signal.update_value(0)
    cpu.do_tick()
    assert output_signal.get_value() == 15

    # Test sum
    control_signal.update_value(4)
    input_signal_b.update_value(1)
    input_signal_a.update_value(1)
    cpu.do_tick()
    assert output_signal.get_value() == 2

    input_signal_b.update_value(5)
    input_signal_a.update_value(7)
    control_signal.update_value(4)
    cpu.do_tick()
    assert output_signal.get_value() == 12
    assert z_signal.get_value() == 0
    assert n_signal.get_value() == 0
    assert o_signal.get_value() == 0
    assert c_signal.get_value() == 0

    # test diff
    input_signal_b.update_value(7)
    input_signal_a.update_value(5)
    control_signal.update_value(5)
    cpu.do_tick()
    assert output_signal.get_value() == 2**16 - 2
    assert z_signal.get_value() == 0
    assert n_signal.get_value() == 1
    assert o_signal.get_value() == 0
    assert c_signal.get_value() == 0

    # test diff
    input_signal_a.update_value(10)
    input_signal_b.update_value(2)
    control_signal.update_value(5)
    cpu.do_tick()
    assert output_signal.get_value() == 8
    assert z_signal.get_value() == 0
    assert n_signal.get_value() == 0
    assert o_signal.get_value() == 0
    assert c_signal.get_value() == 1


def test_hr_signal():
    # test left shift
    input_signal_a.update_value(0b1000000000000000)
    input_signal_hr.update_value(0b0000000000000000)
    control_signal.update_value(0b1010)
    cpu.do_tick()
    assert output_signal.get_value() == 0
    assert hr_signal.get_value() == 1
    assert z_signal.get_value() == 1
    assert n_signal.get_value() == 0
    assert c_signal.get_value() == 0

    # test left shift
    input_signal_a.update_value(0b1000000000000001)
    input_signal_hr.update_value(0b1000000000000010)
    control_signal.update_value(0b1010)
    cpu.do_tick()
    assert output_signal.get_value() == 0b0000000000000010
    assert hr_signal.get_value() == 0b000000000000101
    assert z_signal.get_value() == 0
    assert n_signal.get_value() == 0
    assert c_signal.get_value() == 1

    # test arithmetic right shift
    input_signal_a.update_value(0b1000000000000000)
    input_signal_hr.update_value(0b0000000000000000)
    control_signal.update_value(0b1100)
    cpu.do_tick()
    assert output_signal.get_value() == 0b0100000000000000
    assert hr_signal.get_value() == 0b0000000000000000
    assert z_signal.get_value() == 0
    assert n_signal.get_value() == 0
    assert c_signal.get_value() == 0

    # test arithmetic right shift
    input_signal_a.update_value(0b0000000110000001)
    input_signal_hr.update_value(0b1000000000000001)
    control_signal.update_value(0b1100)
    cpu.do_tick()
    assert output_signal.get_value() == 0b1000000011000000
    assert hr_signal.get_value() == 0b1100000000000000
    assert z_signal.get_value() == 0
    assert n_signal.get_value() == 1
    assert c_signal.get_value() == 1

    # test left rotate
    input_signal_a.update_value(0b1000000000000000)
    input_signal_hr.update_value(0b1000000000000000)
    control_signal.update_value(0b1111)
    cpu.do_tick()
    assert output_signal.get_value() == 0b0000000000000001
    assert hr_signal.get_value() == 0b0000000000000001
    assert z_signal.get_value() == 0
    assert n_signal.get_value() == 0
    assert c_signal.get_value() == 1

    # test left rotate
    input_signal_a.update_value(0b0000000110000001)
    input_signal_hr.update_value(0b0100000000000001)
    control_signal.update_value(0b1111)
    cpu.do_tick()
    assert output_signal.get_value() == 0b0000001100000010
    assert hr_signal.get_value() == 0b1000000000000010
    assert z_signal.get_value() == 0
    assert n_signal.get_value() == 0
    assert c_signal.get_value() == 0
