from simudator.core.processor import Processor
from simudator.core.signal import Signal
from simudator.processor.mia.modules.bus import Bus


def test_bus():
    cpu = Processor()
    in_s_1 = Signal(cpu)
    in_s_2 = Signal(cpu)
    out_s_1 = Signal(cpu)
    out_s_2 = Signal(cpu)
    bus = Bus([in_s_1, in_s_2], [out_s_1, out_s_2])

    in_s_1.add_destination(bus)
    in_s_2.add_destination(bus)
    cpu.add_module(bus)

    in_s_1.set_value(None)
    in_s_2.set_value(None)

    # Test that nothing changes when there is no input to the bus
    cpu.do_tick()
    assert out_s_1.get_value() is None
    assert out_s_2.get_value() is None

    # Test that all output signals are given the input value
    # when there is input to the bus
    in_s_1.update_value(1)
    cpu.do_tick()
    assert out_s_1.get_value() == 1
    assert out_s_2.get_value() == 1

    in_s_1.update_value(None)
    in_s_2.update_value(2)
    cpu.do_tick()
    assert out_s_1.get_value() == 2
    assert out_s_2.get_value() == 2

    # Test that no value is transferred when two input signals
    # are writing to the bus at the same time
    in_s_1.update_value(3)
    in_s_2.update_value(4)
    cpu.do_tick()
    assert out_s_1.get_value() == 2
    assert out_s_2.get_value() == 2
